/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding.mutable;

import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.type.FloatType;

/**
 * Binds FloatType to {@link MutableFloat}
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class MutableFloatBinding extends FloatBinding {

	public static MutableFloatBinding INSTANCE = new MutableFloatBinding();
	
	public MutableFloatBinding() {
		super(FloatType.INSTANCE);
	}
	
	public MutableFloatBinding(FloatType type) {
		super(type);
	}

	@Override
	public Object create(float value) {
		MutableFloat result = new MutableFloat();
		result.value = value;
		return result;
	}

	@Override
	public Object create(Float value) {
		MutableFloat result = new MutableFloat();
		result.value = value;
		return result;
	}

	@Override
	public Object create(Number value) {
		MutableFloat result = new MutableFloat();
		result.value = value.floatValue();
		return result;
	}

	@Override
	public Object create(String value) {
		MutableFloat result = new MutableFloat();
		result.value = Float.valueOf(value);
		return result;
	}

	@Override
	public float getValue_(Object o) throws BindingException {
		MutableFloat result = (MutableFloat) o;
		return result.value;
	}

	@Override
	public Float getValue(Object o) throws BindingException {
		MutableFloat result = (MutableFloat) o;
		return result.value;
	}

	@Override
	public boolean isInstance(Object obj) {
		return obj instanceof MutableFloat;
	}

	@Override
	public void setValue(Object obj, float value) throws BindingException {
		MutableFloat result = (MutableFloat) obj;
		result.value = value;
	}

	@Override
	public void setValue(Object obj, Number value) throws BindingException {
		MutableFloat result = (MutableFloat) obj;
		result.value = value.floatValue();
	}
			
}

