/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.type;

import java.util.Set;

import org.simantics.databoard.DataTypes;
import org.simantics.databoard.annotations.Union;
import org.simantics.databoard.parser.unparsing.DataTypePrinter;
import org.simantics.databoard.util.IdentityHashSet;
import org.simantics.databoard.util.IdentityPair;

@Union({BooleanType.class,
        ByteType.class,
        IntegerType.class,       
        LongType.class,
        FloatType.class,
        DoubleType.class,
        StringType.class,
        RecordType.class,
        ArrayType.class,
        MapType.class,
        OptionalType.class,
        UnionType.class,
        VariantType.class}) 
public abstract class DataType {
	
	protected void collectSubtypes(Set<DataType> subtypes, Set<DataType> recursiveSubtypes) {		
	}
	
	/**
	 * Print the type in data type notation.
	 * 
	 * <a href="https://www.simantics.org/wiki/index.php/Data_type_notation">Datatype Notation</a>
	 * 
	 * See {@link DataTypes#getDataType(String)} to parse string to data type.
	 * 
	 * @return type
	 */
	@Override
	public String toString() {
		return DataTypePrinter.toString(this, true);
	}

	/**
	 * Print the type in data type notation.
	 * 
	 * <a href="https://www.simantics.org/wiki/index.php/Data_type_notation">Datatype Notation</a>
	 * 
	 * See {@link DataTypes#getDataType(String)} to parse string to data type.
	 * 
	 * @return type
	 */
	public String toSingleLineString() {
		return DataTypePrinter.toString(this, false);
	}
	
    public interface Visitor1 {
        void visit(ArrayType b, Object obj);
        void visit(BooleanType b, Object obj);
        void visit(DoubleType b, Object obj);
        void visit(FloatType b, Object obj);
        void visit(IntegerType b, Object obj);
        void visit(ByteType b, Object obj);
        void visit(LongType b, Object obj);
        void visit(OptionalType b, Object obj);
        void visit(RecordType b, Object obj);
        void visit(StringType b, Object obj);
        void visit(UnionType b, Object obj);
        void visit(VariantType b, Object obj);
        void visit(MapType b, Object obj);
    }
	
    public abstract void accept(Visitor1 v, Object obj);
    
    public interface Visitor<T> {
        T visit(ArrayType b);
        T visit(BooleanType b);
        T visit(DoubleType b);
        T visit(FloatType b);
        T visit(IntegerType b);
        T visit(ByteType b);
        T visit(LongType b);
        T visit(OptionalType b);
        T visit(RecordType b);
        T visit(StringType b);
        T visit(UnionType b);
        T visit(VariantType b);
        T visit(MapType b);
    }
    
    public abstract <T> T accept(Visitor<T> v);	

    /**
     * Deep equals-compare
     * 
     * @param obj object to compare with
     * @return true if the object is equal (incl. structurally) with this object
     */
	@Override
	public boolean equals(Object obj) {
		if (this==obj) return true;
		if (!this.getClass().isInstance(obj)) return false;
		
		//if (compareHistory==null)
		//	compareHistory = new HashMap<IdentityPair<DataType, DataType>, Boolean>();
		return deepEquals(obj, null);		
	}
	
// Implemented when enough time
//	public boolean isSubtypeOf(DataType type) {
//		return equals(type);
//	}
    
    protected abstract boolean deepEquals(Object obj, Set<IdentityPair<DataType, DataType>> compareHistory);
    
//    public abstract DataType[] getChildren();
    
}

