/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.tests;

import java.io.IOException;
import java.util.Locale;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.DoubleArrayBinding;
import org.simantics.databoard.parser.repository.DataValueRepository;
import org.simantics.databoard.serialization.BinarySerializationFormat;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.serialization.SerializerConstructionException;

public class TestSomething {

    static class LocalizedText {

        public static final Locale FINNISH = new Locale("fi");
        
        public static LocalizedText english(String text) {
            return new LocalizedText(text, Locale.ENGLISH.toString());
        }
        
        public static LocalizedText finnish(String text) {
            return new LocalizedText(text, FINNISH.toString());
        }
        
        public LocalizedText() {}
        
        public LocalizedText(String text, String locale) {
            this.text = text;
            this.locale = locale;
        }
        
        public String text;
        public String locale;

        @Override
        public String toString() {
            return text;
        }
        
    }

    public static void main(String[] args) throws SerializationException, BindingException, IOException, SerializerConstructionException {
        Binding dab = DoubleArrayBinding.INSTANCE;
        double[] array = new double[5];
        Serializer ser = dab.serializer(BinarySerializationFormat.INSTANCE);
        byte[] dada = ser.serialize(array);
        array = (double[]) ser.deserialize(dada);
        System.out.println(array);
        
        LocalizedText lt = LocalizedText.english("root");
        Binding b = Bindings.getBindingUnchecked(LocalizedText.class);        
        Serializer s = b.serializer(BinarySerializationFormat.INSTANCE);
        
        DataValueRepository vr = new DataValueRepository(); 

        b.printValue(lt, System.out, vr, false);
        
        Binding lb = Bindings.getBindingUnchecked(Locale.class);
        lb.printValue(Locale.UK, System.out, vr, true);                
        
        byte[] data = s.serialize(lt);
        LocalizedText lt2 = (org.simantics.databoard.tests.TestSomething.LocalizedText) s.deserialize(data);
        
        
        
    }
    
}

